
from colmet.metrics.base import get_counters_class


def test_get_counters_class():
    '''Test: import the taskstats_default metric class'''
    counter_class = get_counters_class('taskstats_default')
    assert(counter_class._get_metric_name() == 'taskstats_default')


def setup_metric():
    global counter
    global timestamp

    counter_class = get_counters_class('taskstats_default')
    counter = counter_class()


def test_counters_initialisation():
    '''Test: taskstats_metric initialization'''
    setup_metric()



"""This package contains class for representing colmet data"""
import time
import os
import socket
import logging

LOG = logging.getLogger()

from .exceptions import NoJobFoundError


class Info(object):
    '''
    Base class for Information class that do the common initialization
    '''
    def __init__(self, input_backend):
        self.input_backend = input_backend
        self.counters_class = input_backend.get_counters_class()
        self.stats_total = self.counters_class()
        self.stats_delta = None

    def get_stats(self):
        return self.stats_total

    def update_stats(self, timestamp, job_id, hostname):
        self.stats_total.timestamp = timestamp
        self.stats_total.job_id = job_id
        self.stats_total.hostname = hostname


class TaskInfo(Info):
    '''
    Represent Information corresponding to a task
    '''
    def __init__(self, tid, input_backend):
        Info.__init__(self, input_backend)
        self.tid = tid
        self.mark = True
        self.backend_request = self.counters_class.build_request(input_backend,
                                                                 tid)

    def update_stats(self, timestamp, job_id, hostname):
        '''
        Update the current metrics of the task
        '''
        stats = self.counters_class.fetch(self.input_backend,
                                          self.backend_request)
        if stats:
            if not self.stats_delta:
                self.stats_total = stats
                self.stats_delta = stats
            stats.delta(self.stats_total, self.stats_delta)
            self.stats_total = stats
            self.mark = False
        Info.update_stats(self, timestamp, job_id, hostname)


class ProcessInfo(Info):
    '''
    Representing the information corresponding to a process
    '''
    def __init__(self, pid, input_backend):
        Info.__init__(self, input_backend)
        self.tgid = pid
        self.tasks = {}
        self.get_task(pid)
        self.mark = True

    def update_stats(self, timestamp, job_id, hostname):
        '''
        Update the current metrics of the process
        '''
        for tid in self.list_tids():
            task = self.get_task(tid)
            task.update_stats(timestamp, job_id, hostname)

        stats_delta = self.counters_class()
        stats_delta.type = 'process'
        stats_delta.id = self.tgid

        for tid, task in self.tasks.items():
            if task.mark:
                msg = (
                    "Task %s no more exists."
                    "Removing from the Job %s"
                    % (tid, self.tgid)
                )
                LOG.debug(msg)
                del self.tasks[tid]
            else:
                task.mark = True
                stats_delta.accumulate(task.stats_delta, stats_delta)

        self.stats_delta = stats_delta
        if not self.stats_total:
            self.stats_total = stats_delta
        else:
            self.stats_total.accumulate(stats_delta, self.stats_total)

        self.mark = False
        Info.update_stats(self, timestamp, job_id, hostname)

        return True

    def get_process(self, tgid):
        '''
        Retrieve the process corresponding to the tgid
        '''
        if tgid != self.tgid:
            return None
        return self

    def list_tids(self):
        '''
        Retrieve the list of tasks of the current process
        '''
        try:
            tids = map(int, os.listdir('/proc/%d/task' % self.tgid))
        except OSError:
            self.mark = True
            return []

        return tids

    def get_task(self, tid):
        '''
        Retrieve the task corresponding to the task id 'tid'
        '''
        task = self.tasks.get(tid, None)
        if not task:
            task = TaskInfo(tid, self.input_backend)
            self.tasks[tid] = task
        return task


class CGroupInfo(Info):
    '''
    Represent the information corresponding to a cgroup
    '''
    def __init__(self, cgroup_path, input_backend):
        Info.__init__(self, input_backend)
        self.cgroup_path = cgroup_path
        self.tasks = {}

    def list_tids(self):
        '''
        Return the list of tasks of the current process
        '''
        try:
            f_tasks = open(os.path.join(self.cgroup_path, 'tasks'), 'r')
            pids = map(int, f_tasks.read().split())
            f_tasks.close()
        except OSError:
            return []
        except IOError:
            return []
        return pids

    def get_task(self, tid):
        '''
        Return the task corresponding to the given tid
        '''
        task = self.tasks.get(tid, None)
        if not task:
            task = TaskInfo(tid, self.input_backend)
            self.tasks[tid] = task
        return task

    def update_stats(self, timestamp, job_id, hostname):
        '''
        Update the current metrics for this cgroup.
        '''

        stats_delta = self.counters_class()
        stats_delta.type = 'cgroup'
        stats_delta.id = self.cgroup_path

        if len(self.list_tids()) > 0:
            for tid in self.list_tids():
                task = self.get_task(tid)
                task.update_stats(timestamp, job_id, hostname)

            for tid, task in self.tasks.items():
                if task.mark:
                    msg = (
                        "Process/Task %s no more exists."
                        "Removing from the Job %s"
                        % (tid, self.cgroup_path)
                    )
                    LOG.debug(msg)
                    del self.tasks[tid]
                else:
                    task.mark = True
                    stats_delta.accumulate(task.stats_delta, stats_delta)

            self.stats_delta = stats_delta
            self.stats_total.accumulate(stats_delta, self.stats_total)

            self.void_cpuset = False

        else:  # no task in this cgroup....
            LOG.info("no taks in this cgroup")
            # raise VoidCpusetError

        Info.update_stats(self, timestamp, job_id, hostname)
        return True


class ProcStatsInfo(Info):
    '''
    Represent the information corresponding to a node (not only one job)
    '''
    def __init__(self, input_backend):
        Info.__init__(self, input_backend)
        self.input_backend = self.input_backend

    def update_stats(self, timestamp, job_id, hostname):
        '''
        Update the current metrics for the node (job_id=0)
        '''
        stats = self.counters_class.fetch(self.input_backend)
        if stats:
            if not self.stats_delta:
                self.stats_total = stats
                self.stats_delta = stats
            stats.delta(self.stats_total, self.stats_delta)
            self.stats_total = stats

        Info.update_stats(self, timestamp, job_id, hostname)

        return True


class Job(object):
    '''
    Represent the Job to monitor
    '''
    def __init__(self, input_backend, job_id, options):
        self.job_children = []
        self.input_backend = input_backend
        self.options = options
        self.timestamp = int(time.time())
        self.duration = None
        self.job_id = job_id
        self._hostname = socket.gethostname()
        # void_cpuset is used to avoid sending null stats to output backends
        # TODO rm
        self.void_cpuset = True

        self.job_children.extend(
            [TaskInfo(tid, input_backend) for tid in options.tids]
        )
        self.job_children.extend(
            [ProcessInfo(pid, input_backend) for pid in options.pids]
        )
        self.job_children.extend(
            [CGroupInfo(cgroup, input_backend) for cgroup in options.cgroups]
        )

        # to monitor global node resources activities
        if job_id == 0:
            self.job_children.extend([ProcStatsInfo(input_backend)])

        number_of_child = len(self.job_children)
        if number_of_child == 0:
            raise NoJobFoundError
        msg = "the job contains %s items" % (number_of_child)
        LOG.info(msg)

        self.update_stats()

    def list_tgids(self):
        '''
        Return the list of process id corresponding to the running jobs
        '''

        tgids = []
        for item in self.job_children:
            tgids.extend(item.list_tgids())

        current_tgids = Job.list_running_tgids()

        return list(set(current_tgids).intersection(set(tgids)))

    @staticmethod
    def list_running_tgids():
        '''
        Return the list of process id currently running
        '''
        proc_ls = os.listdir('/proc')
        return [int(tgid) for tgid in proc_ls if '0' <= tgid[0] <= '9']

    def update_stats(self):
        '''
        Update the metrics for all the job's children
        '''
        new_timestamp = int(time.time())
        self.duration = new_timestamp - self.timestamp
        self.timestamp = new_timestamp

        for item in self.job_children:
            item.update_stats(self.timestamp, self.job_id, self._hostname)

    def get_children(self):
        '''
        Return the job lists
        '''
        return self.job_children

    def get_stats(self):
        return [child.get_stats() for child in self.job_children]



'''
ZeroMQ input backend for colmet collector
'''
import zmq
from zmq.eventloop import ioloop, zmqstream
import logging
import socket
import struct


LOG = logging.getLogger()

from colmet.common.metrics.base import BaseCounters

from colmet.common.backends.base import InputBaseBackend, OutputBaseBackend

try:
    _snd_hwm = zmq.HWM
    _rcv_hwm = zmq.HWM
except AttributeError:
    _snd_hwm = zmq.SNDHWM
    _rcv_hwm = zmq.RCVHWM


class ZMQInputBackend(InputBaseBackend):
    __backend_name__ = "zeromq"

    def open(self):
        self.context = zmq.Context()
        self.loop = ioloop.IOLoop.instance()
        self.socket = self.context.socket(zmq.PULL)
        self.socket.setsockopt(zmq.LINGER, self.options.zeromq_linger)
        self.socket.setsockopt(_rcv_hwm, self.options.zeromq_hwm)
        LOG.debug("Use the bind URI '%s'" % self.options.zeromq_bind_uri)
        self.socket.bind(self.options.zeromq_bind_uri)
        self.stream = zmqstream.ZMQStream(self.socket, self.loop)

    def close(self):
        self.socket.close()
        self.context.term()

    def unpack(self, msgs, callback):
        counters_list = []
        for raw in msgs:
            try:
                counters_list.extend(BaseCounters.unpack_to_list(raw))
            except Exception as e:
                LOG.exception(e)
        LOG.debug("%s counters received" % len(counters_list))
        for counters in counters_list:
            callback(counters)

    def on_recv(self, callback):
        intern_callback = lambda x: self.unpack(x, callback)
        self.stream.on_recv(intern_callback)


class ZMQOutputBackend(OutputBaseBackend):
    '''
    zmq node backend class
    '''
    __backend_name__ = "zeromq"

    def open(self):
        self.context = zmq.Context()
        self.hostname = socket.gethostname()
        self.socket = self.context.socket(zmq.PUSH)
        self.socket.setsockopt(zmq.LINGER, self.options.zeromq_linger)
        self.socket.setsockopt(_snd_hwm, self.options.zeromq_hwm)
        self.socket.connect(self.options.zeromq_uri)
        LOG.debug("Use the URI '%s'" % self.options.zeromq_uri)

    def close(self):
        self.socket.close()
        self.context.term()

    def push(self, counters_list):
        '''
        push the metrics to the output backend
        '''
        if len(counters_list) > 0:
            try:
                raw = BaseCounters.pack_from_list(counters_list)
                self.socket.send(raw)
            except struct.error as e:
                LOG.error("An error occurred during packet creation : %s" % e)



import logging

LOG = logging.getLogger()

from .base import UInt64, UFloat, BaseCounters


class ProcstatsCounters(BaseCounters):
    __metric_name__ = 'procstats_default'

    counters_procstats = {
        # 'key': ( offset,length, type, repr, acc )
        'uptime_total': (UInt64(), 'sec', 'none', 'uptime.total'),
        'uptime_idle': (UInt64(), 'sec', 'none', 'uptime.idle'),
        'meminfo_memtotal': (UInt64(), 'kbytes', 'none', 'meminfo_memtotal'),
        'meminfo_memfree': (UInt64(), 'kbytes', 'none', 'meminfo_memfree'),
        'meminfo_buffers': (UInt64(), 'kbytes', 'none', 'meminfo_buffers'),
        'meminfo_cached': (UInt64(), 'kbytes', 'none', 'meminfo_cached'),
        'meminfo_swapcached': (UInt64(), 'kbytes', 'none',
                               'meminfo_swapcached'),
        'meminfo_active': (UInt64(), 'kbytes', 'none', 'meminfo_active'),
        'meminfo_inactive': (UInt64(), 'kbytes', 'none', 'meminfo_inactive'),
        'meminfo_unevictable': (UInt64(), 'kbytes', 'none',
                                'meminfo_unevictable'),
        'meminfo_mlocked': (UInt64(), 'kbytes', 'none', 'meminfo_mlocked'),
        'meminfo_swaptotal': (UInt64(), 'kbytes', 'none', 'meminfo_swaptotal'),
        'meminfo_swapfree': (UInt64(), 'kbytes', 'none', 'meminfo_swapfree'),
        'meminfo_dirty': (UInt64(), 'kbytes', 'none', 'meminfo_dirty'),
        'meminfo_writeback': (UInt64(), 'kbytes', 'none', 'meminfo_writeback'),
        'meminfo_anonpages': (UInt64(), 'kbytes', 'none', 'meminfo_anonpages'),
        'meminfo_mapped': (UInt64(), 'kbytes', 'none', 'meminfo_mapped'),
        'meminfo_shmem': (UInt64(), 'kbytes', 'none', 'meminfo_shmem'),
        'meminfo_slab': (UInt64(), 'kbytes', 'none', 'meminfo_slab'),
        'meminfo_sreclaimable': (UInt64(), 'kbytes', 'none',
                                 'meminfo_sreclaimable'),
        'meminfo_sunreclaim': (UInt64(), 'kbytes', 'none',
                               'meminfo_sunreclaim'),
        'meminfo_kernelstack': (UInt64(), 'kbytes', 'none',
                                'meminfo_kernelstack'),
        'meminfo_pagetables': (UInt64(), 'kbytes', 'none',
                               'meminfo_pagetables'),
        'meminfo_nfs_unstable': (UInt64(), 'kbytes', 'none',
                                 'meminfo_nfs_unstable'),
        'meminfo_bounce': (UInt64(), 'kbytes', 'none', 'meminfo_bounce'),
        'meminfo_writebacktmp': (UInt64(), 'kbytes', 'none',
                                 'meminfo_writebacktmp'),
        'meminfo_commitlimit': (UInt64(), 'kbytes', 'none',
                                'meminfo_commitlimit'),
        'meminfo_committed_as': (UInt64(), 'kbytes', 'none',
                                 'meminfo_committed_as'),
        'meminfo_vmalloctotal': (UInt64(), 'kbytes', 'none',
                                 'meminfo_vmalloctotal'),
        'meminfo_vmallocused': (UInt64(), 'kbytes', 'none',
                                'meminfo_vmallocused'),
        'meminfo_vmallocchunk': (UInt64(), 'kbytes', 'none',
                                 'meminfo_vmallocchunk'),
        'meminfo_hardwarecorrupted': (UInt64(), 'kbytes', 'none',
                                      'meminfo_hardwarecorrupted'),
        'meminfo_anonhugepages': (UInt64(), 'kbytes', 'none',
                                  'meminfo_anonhugepages'),
        'meminfo_hugepages_total': (UInt64(), 'count', 'none',
                                    'meminfo_hugepages_total'),
        'meminfo_hugepages_free': (UInt64(), 'count', 'none',
                                   'meminfo_hugepages_free'),
        'meminfo_hugepages_rsvd': (UInt64(), 'count', 'none',
                                   'meminfo_hugepages_rsvd'),
        'meminfo_hugepages_surp': (UInt64(), 'count', 'none',
                                   'meminfo_hugepages_surp'),
        'meminfo_hugepagesize': (UInt64(), 'count', 'none',
                                 'meminfo_hugepagesize'),
        'meminfo_directmap4k': (UInt64(), 'kbytes', 'none',
                                'meminfo_directmap4k'),
        'meminfo_directmap2m': (UInt64(), 'kbytes', 'none',
                                'meminfo_directmap2m'),
        'vmstat_pgpgin': (UInt64(), 'count', 'none', 'vmstat_pgpgin'),
        'vmstat_pgpgout': (UInt64(), 'count', 'none', 'vmstat_pgpgout'),
        'vmstat_pswpin': (UInt64(), 'count', 'none', 'vmstat_pswpin'),
        'vmstat_pswpout': (UInt64(), 'count', 'none', 'vmstat_pswpout'),
        'vmstat_pgfault': (UInt64(), 'count', 'none', 'vmstat_pgfault'),
        'vmstat_pgmajfault': (UInt64(), 'count', 'none', 'vmstat_pgmajfault'),
        'stat_cpu_user': (UInt64(), 'n/a', 'none', 'stat_cpu_user'),
        'stat_cpu_nice': (UInt64(), 'n/a', 'none', 'stat_cpu_nice'),
        'stat_cpu_system': (UInt64(), 'n/a', 'none', 'stat_cpu_nice'),
        'stat_cpu_idle': (UInt64(), 'n/a', 'none', 'stat_cpu_nice'),
        'stat_cpu_iowait': (UInt64(), 'n/a', 'none', 'stat_cpu_nice'),
        'stat_cpu_irq': (UInt64(), 'n/a', 'none', 'stat_cpu_nice'),
        'stat_cpu_softirq': (UInt64(), 'n/a', 'none', 'stat_cpu_nice'),
        'stat_cpu_guest': (UInt64(), 'n/a', 'none', 'stat_cpu_nice'),
        'stat_cpu_guest_nice': (UInt64(), 'n/a', 'none', 'stat_cpu_nice'),
        'stat_intr': (UInt64(), 'n/a', 'none', 'stat_intr'),
        'stat_ctxt': (UInt64(), 'n/a', 'none', 'stat_ctxt'),
        'stat_processes': (UInt64(), 'n/a', 'none', 'stat_processes'),
        'stat_procs_blocked': (UInt64(), 'n/a', 'none', 'stat_procs_blocked'),
        'loadavg_1min': (UFloat(), 'n/a', 'none', 'loadavg_1min'),
        'loadavg_5min': (UFloat(), 'n/a', 'none', 'loadavg_5min'),
        'loadavg_15min': (UFloat(), 'n/a', 'none', 'loadavg_15min'),
        'loadavg_runnable': (UFloat(), 'n/a', 'none', 'loadavg_runnable'),
        'loadavg_total_threads': (UFloat(), 'n/a', 'none',
                                  'loadavg_total_threads')
    }

    counters_procstats_to_get = [
        'uptime_total',
        'uptime_idle',
        'meminfo_memtotal',
        'meminfo_memfree',
        'meminfo_buffers',
        'meminfo_cached',
        'meminfo_swapcached',
        'meminfo_active',
        'meminfo_inactive',
        'meminfo_unevictable',
        'meminfo_mlocked',
        'meminfo_swaptotal',
        'meminfo_swapfree',
        'meminfo_dirty',
        'meminfo_writeback',
        'meminfo_anonpages',
        'meminfo_mapped',
        'meminfo_shmem',
        'meminfo_slab',
        'meminfo_sreclaimable',
        'meminfo_sunreclaim',
        'meminfo_kernelstack',
        'meminfo_pagetables',
        'meminfo_nfs_unstable',
        'meminfo_bounce',
        'meminfo_writebacktmp',
        'meminfo_commitlimit',
        'meminfo_committed_as',
        'meminfo_vmalloctotal',
        'meminfo_vmallocused',
        'meminfo_vmallocchunk',
        'meminfo_hardwarecorrupted',
        'meminfo_anonhugepages',
        'meminfo_hugepages_total',
        'meminfo_hugepages_free',
        'meminfo_hugepages_rsvd',
        'meminfo_hugepages_surp',
        'meminfo_hugepagesize',
        'meminfo_directmap4k',
        'meminfo_directmap2m',
        'vmstat_pgpgin',
        'vmstat_pgpgout',
        'vmstat_pswpin',
        'vmstat_pswpout',
        'vmstat_pgfault',
        'vmstat_pgmajfault',
        'stat_cpu_user',
        'stat_cpu_nice',
        'stat_cpu_system',
        'stat_cpu_idle',
        'stat_cpu_iowait',
        'stat_cpu_irq',
        'stat_cpu_softirq',
        'stat_cpu_guest',
        'stat_cpu_guest_nice',
        'stat_intr',
        'stat_ctxt',
        'stat_processes',
        'stat_procs_blocked',
        'loadavg_1min',
        'loadavg_5min',
        'loadavg_15min',
        'loadavg_runnable',
        'loadavg_total_threads'
        # 'sys_numa_zoneallocs',
        # 'sys_numa_zoneallocs',
        # 'sys_numa_foreign_allocs',
        # 'sys_numa_allocation',
        # 'sys_numa_allocation',
        # 'sys_numa_interleave',
    ]

    _counters = []
    for c_name in counters_procstats_to_get:
        (c_type, c_repr, c_acc, c_descr) = counters_procstats[c_name]
        _counters.append((c_name, c_type, c_repr, c_acc, c_descr))

    @classmethod
    def get_zero_counters(cls):
        return cls(procstats_buffer=None, raw=None)

    @classmethod
    def fetch(cls, procstats_backend):
        return procstats_backend.get_procstats()

    def __init__(self, procstats_buffer=None, raw=None):
        BaseCounters.__init__(self, raw=raw)
        if raw is not None:
            pass
        elif procstats_buffer is None:
            self._empty_fill()
        else:
            for name in ProcstatsCounters._counter_definitions:
                self._counter_values[name] = procstats_buffer[name]



'''
Colmet-node User Interface
'''
import logging
import argparse
import signal
import sys
import time

from colmet import VERSION
from colmet.node.backends.procstats import ProcstatsBackend
from colmet.node.backends.taskstats import TaskstatsBackend
from colmet.common.backends.zeromq import ZMQOutputBackend
from colmet.common.utils import AsyncFileNotifier, as_thread
from colmet.common.exceptions import Error, NoneValueError


LOG = logging.getLogger()


class Task(object):

    def __init__(self, name, options):
        self.name = name
        self.options = options
        self.input_backends = []
        self.taskstats_backend = TaskstatsBackend(self.options)
        self.input_backends.append(self.taskstats_backend)
        if not self.options.disable_procstats:
            self.input_backends.append(ProcstatsBackend(self.options))
        self.zeromq_output_backend = ZMQOutputBackend(self.options)

    @as_thread
    def check_jobs_thread(self):
        notifier = \
            AsyncFileNotifier(paths=self.options.cpuset_rootpath,
                              callback=self.taskstats_backend.update_job_list)
        # Initial job list update
        self.update_job_list()
        notifier.loop()

    def update_job_list(self):
        self.taskstats_backend.update_job_list()

    def start(self):
        LOG.info("Starting %s" % self.name)
        for backend in self.input_backends:
            backend.open()
        self.zeromq_output_backend.open()
        signal.signal(signal.SIGINT, self.terminate)
        signal.signal(signal.SIGTERM, self.terminate)
        self.check_jobs_thread.start()
        self.loop()

    def terminate(self, signum, frame):
        LOG.info("Received a signal (%d)" % signum)
        LOG.info("Terminating %s properly..." % self.name)
        self.zeromq_output_backend.close()
        sys.exit(0)

    def sleep(self):
        now = time.time()
        time_towait = (((now // self.options.sampling_period) + 1) *
                       self.options.sampling_period - now)
        time.sleep(time_towait)

    def loop(self):
        while True:
            now = time.time()
            LOG.debug("Gathering the metrics")
            counters_list = []
            for backend in self.input_backends:
                pulled_counters = backend.pull()

                if backend.get_backend_name() == 'taskstats':
                    if len(pulled_counters) > 0:
                        for counters in pulled_counters:
                            counters_list += counters
                else:
                    counters_list += pulled_counters

                LOG.debug("%s metrics has been pulled width %s" %
                          (len(pulled_counters), backend.get_backend_name()))

            LOG.debug("time to take measure: %s sec" % (time.time() - now))

            if len(counters_list) > 0:
                # print "nb counters_list", len(counters_list)
                try:
                    self.zeromq_output_backend.push(counters_list)
                    LOG.debug("%s metrics has been pushed with zeromq"
                              % (len(counters_list)))
                except (NoneValueError, TypeError):
                    LOG.debug("Values for metrics are not there.")
            # sleep to next sampling
            self.sleep()


#
# Main program
#

DESCRIPTION = '''Display and/or collect cpu, memory and i/o bandwidth used
by the processes in a cpuset or a cgroup.'''


def main():
    formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(description=DESCRIPTION,
                                     formatter_class=formatter)

    parser.add_argument('--version', action='version',
                        version='colmet version %s' % VERSION)

    parser.add_argument('-v', '--verbose', action='count', dest="verbosity",
                        default=1)

    parser.add_argument('-s', '--sample-period', type=float,
                        dest='sampling_period', default=5,
                        help='Sampling period of measuring in seconds')

    parser.add_argument('--disable-procstats', action="store_true",
                        default=False, dest="disable_procstats",
                        help='Disables node monitoring based on some /proc '
                             'subdirectories contents. Measures are '
                             'associated to the fictive job with 0 as '
                             'identifier (job_id)')

    group = parser.add_argument_group('Taskstat')

    group.add_argument('-c', '--cgroup', dest='cgroups',
                       action='append', default=[],
                       help='cgroup/cpuset to monitor', metavar='CGROUP')
    group.add_argument('-p', '--pid', type=int, dest='pids',
                       action='append', default=[],
                       help='process id to monitor', metavar='PID')
    group.add_argument('-t', '--tid', type=int, dest='tids',
                       action='append', default=[],
                       help='task ids to monitor', metavar='TID')
    group.add_argument('--cpuset_rootpath', dest='cpuset_rootpath',
                       action='append', default=["/dev/oar_cgroups/oar/"],
                       help='cpuset root path', metavar='CPUSETROOTPATH')
    group.add_argument('--regex_job_id', dest='regex_job_id',
                       action='append', default=['_(\d+)$'],
                       metavar='REGEXJOBID',
                       help='regular expression to extract job_id from '
                            'cpuset directory name', )
    parser.add_argument_group(group)

    group = parser.add_argument_group('Zeromq')

    group.add_argument("--zeromq-uri", dest='zeromq_uri',
                       help="ZeroMQ collector URI",
                       default='tcp://127.0.0.1:5556')

    group.add_argument("--zeromq-hwm", type=int,
                       default=1000, dest='zeromq_hwm',
                       help="The high water mark is a hard limit on the"
                            " maximum number of outstanding messages "
                            " ZeroMQ shall queue in memory. The value "
                            " of zero means \"no limit\".")

    group.add_argument("--zeromq-linger", type=int,
                       default=0, dest='zeromq_linger',
                       help="Set the linger period for the specified socket."
                            " The value of -1 specifies an infinite linger"
                            " period. The value of 0 specifies no linger"
                            " period.  Positive values specify an upper bound"
                            " for the  linger period in milliseconds.")

    parser.add_argument_group(group)

    args = parser.parse_args()

    # Set the logging value (always display CRITICAL and ERROR)
    logging.basicConfig(
        format='%(asctime)s - %(levelname)s - %(message)s',
        datefmt='%d/%m/%Y %H:%M:%S',
        level=40 - args.verbosity * 10,
    )

    # run
    try:
        Task(sys.argv[0], args).start()
    except KeyboardInterrupt:
        pass
    except Error as err:
        err.show()
        sys.exit(1)
    except Exception as err:
        msg = "Error not handled '%s'" % err.__class__.__name__
        logging.critical(msg)
        if logging.getLogger().getEffectiveLevel() <= logging.DEBUG:
            print(repr(err))
        raise


if __name__ == '__main__':
    main()



'''
Colmet Exceptions
'''
import logging
import sys
import traceback

LOG = logging.getLogger()


class Error(Exception):
    '''
    Generic error
    '''
    desc = "Generic Error"

    def __init__(self, desc_args={}):
        super(Error, self).__init__()
        self.exc_type, self.exc_value, self.exc_traceback = sys.exc_info()
        self.desc_args = desc_args

    def format(self):
        """
        Return the formatted error
        """
        msg = self.desc % self.desc_args
        if LOG.getEffectiveLevel() <= logging.DEBUG:
            msg += "\n\n" + traceback.format_exc(self.exc_traceback)
        return msg

    def show(self):
        '''
        Display the formatted error
        '''
        LOG.error(self.format())


class NoJobFoundError(Error):
    '''
    No Job Found Error
    '''
    desc = (
        "No job found/specified. "
        "You probably need to provide information "
        "about job to monitor on the command line."
    )


class NoEnoughPrivilegeError(Error):
    '''
    No Enough Privileges Error
    '''
    desc = (
        "No enough privileges to run this program. "
        "You probably need to be root."
    )


class UnableToFindCounterClassError(Error):
    '''
    Unable to find a counter class Error
    '''
    desc = (
        "An error occurs when importing the class %s:"
    )

    def format(self):
        msg = Error.format(self)
        if self.error is not None:
            msg += "\n" + str(self.error)
        return msg

    def __init__(self, class_name, error=None):
        Error.__init__(self)
        self.desc_args = class_name
        self.error = error


class CounterAlreadyExistError(Error):
    '''
    The counter Already Exist Error
    '''
    desc = (
        "A counter has been registered twice. "
        "This shouldn't append. Please contact the developpers."
    )


class JobNeedToBeDefinedError(Error):
    '''
    The monitored job id must be defined Error
    '''
    desc = (
        "Unable to find jobs in this node. "
        "Please check Taskstat options ('-h')."
    )


class UnableToFindLibraryError(Error):
    '''
    Unable to find library error
    '''
    desc = (
        "Unable to load the library '%s'. "
        "\nPlease check that your environnement provide it."
    )


class MultipleBackendsNotSupported(Error):
    '''
    Colmet doesn't support multiple input/output backend in the same process
    '''
    desc = (
        "Colmet doesn't support multiple "
        "input/output backend in the same process."
    )


class NotEnoughInputBackend(Error):
    desc = (
        "You need to provide at least one input backend "
        "on the command line"
    )


class FileAlreadyOpenWithDifferentModeError(Error):
    desc = (
        "The file %s is already opened with a different "
        "filemode access in the same process. This is "
        "not supported."
    )


class TimeoutException(Exception):
    '''
    A simple class to handle timeout for the main colmet loop
    '''
    pass


class NoneValueError(Error):
    '''
    Unable to retreive value for this metric
    '''
    pass


class VoidCpusetError(Error):
    '''
    This Cpuset is void
    '''
    pass



import os
import re
import errno
import struct
import copy
import logging

from colmet.common.metrics.taskstats import TaskstatsCounters
from colmet.common.exceptions import (NoEnoughPrivilegeError,
                                      JobNeedToBeDefinedError)
from colmet.common.backends.base import InputBaseBackend
from colmet.common.job import Job

LOG = logging.getLogger()


class TaskstatsBackend(InputBaseBackend):

    __backend_name__ = "taskstats"

    def open(self):
        self.jobs = {}

        self.taskstats_nl = TaskStatsNetlink(self.options)

        if len(self.job_id_list) < 1 \
                and self.options.cpuset_rootpath == []:
            raise JobNeedToBeDefinedError()
        if len(self.job_id_list) == 1:
            job_id = self.job_id_list[0]
            self.jobs[job_id] = Job(self, job_id, self.options)
        else:
            for i, job_id in enumerate(self.job_id_list):
                self.jobs[job_id] = Job(self, job_id, self.options)

    def close(self):
        pass

    def build_request(self, pid):
        return self.taskstats_nl.build_request(pid)

    def get_task_stats(self, request):
        counters = self.taskstats_nl.get_single_task_stats(request)
        return counters

    def pull(self):
        for job in self.jobs.values():
            LOG.debug("pull job :" + str(job.job_id))
            job.update_stats()
        return [job.get_stats() for job in self.jobs.values()]

    def get_counters_class(self):
        return TaskstatsCounters

    def create_options_job_cgroups(self, cgroups):
        # options are duplicated to allow modification per jobs, here
        # cgroups parametter
        options = copy.copy(self.options)
        options.cgroups = cgroups
        return options

    def update_job_list(self):
        """Used to maintained job list upto date by adding new jobs and
        removing ones to monitor accordingly to cpuset_rootpath and
        regex_job_id.
        """
        cpuset_rootpath = self.options.cpuset_rootpath[0]
        regex_job_id = self.options.regex_job_id[0]

        job_ids = set([])
        filenames = {}
        for filename in os.listdir(cpuset_rootpath):
            jid = re.findall(regex_job_id, filename)
            if len(jid) > 0:
                job_ids.add(jid[0])
                filenames[jid[0]] = filename

        monitored_job_ids = set(self.job_id_list)
        # Add new jobs
        for job_id in (job_ids - monitored_job_ids):
            job_path = cpuset_rootpath + "/" + filenames[job_id]
            options = self.create_options_job_cgroups([job_path])
            self.jobs[job_id] = Job(self, int(job_id), options)
        # Del ended jobs

        for job_id in (monitored_job_ids - job_ids):
            del self.jobs[job_id]
        # udpate job_id list to monitor
        self.job_id_list = list(job_ids)

#
# Taskstats Netlink
#

from .genetlink.netlink import Connection, NETLINK_GENERIC, U32Attr, \
    NLM_F_REQUEST
from .genetlink.genetlink import Controller, GeNlMessage

TASKSTATS_CMD_GET = 1

TASKSTATS_CMD_ATTR_PID = 1
TASKSTATS_CMD_ATTR_TGID = 2

TASKSTATS_TYPE_PID = 1
TASKSTATS_TYPE_TGID = 2
TASKSTATS_TYPE_STATS = 3
TASKSTATS_TYPE_AGGR_PID = 4
TASKSTATS_TYPE_AGGR_TGID = 5


class TaskStatsNetlink(object):
    # Keep in sync with format_stats() and pinfo.did_some_io()

    def __init__(self, options):
        self.options = options
        self.connection = Connection(NETLINK_GENERIC)
        controller = Controller(self.connection)
        self.family_id = controller.get_family_id('TASKSTATS')

    def build_request(self, tid):
        return GeNlMessage(self.family_id, cmd=TASKSTATS_CMD_GET,
                           attrs=[U32Attr(TASKSTATS_CMD_ATTR_PID, tid)],
                           flags=NLM_F_REQUEST)

    def get_single_task_stats(self, request):
        request.send(self.connection)
        try:
            reply = GeNlMessage.recv(self.connection)
        except OSError as e:
            if e.errno == errno.ESRCH:
                # OSError: Netlink error: No such process (3)
                return
            if e.errno == errno.EPERM:
                raise NoEnoughPrivilegeError
            raise
        for attr_type, attr_value in reply.attrs.iteritems():
            if attr_type == TASKSTATS_TYPE_AGGR_PID:
                reply = attr_value.nested()
                break
            # elif attr_type == TASKSTATS_TYPE_PID:
            #    pass
        else:
            return
        taskstats_data = reply[TASKSTATS_TYPE_STATS].data
        if len(taskstats_data) < 272:
            # Short reply
            return
        taskstats_version = struct.unpack('H', taskstats_data[:2])[0]
        assert taskstats_version >= 4
        return TaskstatsCounters(taskstats_buffer=taskstats_data)



'''
stdout backend : print information to stdout
'''
import logging
import os

from colmet.common.exceptions import FileAlreadyOpenWithDifferentModeError
from colmet.common.metrics import get_counters_class
from colmet.common.backends.base import OutputBaseBackend
import tables


LOG = logging.getLogger()

HDF5_BACKEND_VERSION = 2


class HDF5TaskstatsCounters(object):
    Counters = get_counters_class("taskstats_default")

    class HDF5TableDescription(tables.IsDescription):

        metric_backend = tables.StringCol(255)
        timestamp = tables.Int64Col(dflt=-1)
        hostname = tables.StringCol(255)
        job_id = tables.Int64Col(dflt=-1)

        cpu_count = tables.Int64Col(dflt=-1)
        cpu_delay_total = tables.Int64Col(dflt=-1)
        blkio_count = tables.Int64Col(dflt=-1)
        blkio_delay_total = tables.Int64Col(dflt=-1)
        swapin_count = tables.Int64Col(dflt=-1)
        swapin_delay_total = tables.Int64Col(dflt=-1)
        cpu_run_real_total = tables.Int64Col(dflt=-1)
        cpu_run_virtual_total = tables.Int64Col(dflt=-1)
        ac_btime = tables.Int64Col(dflt=-1)
        ac_etime = tables.Int64Col(dflt=-1)
        ac_utime = tables.Int64Col(dflt=-1)
        ac_stime = tables.Int64Col(dflt=-1)
        ac_minflt = tables.Int64Col(dflt=-1)
        ac_majflt = tables.Int64Col(dflt=-1)
        coremem = tables.Int64Col(dflt=-1)
        virtmem = tables.Int64Col(dflt=-1)
        read_char = tables.Int64Col(dflt=-1)
        write_char = tables.Int64Col(dflt=-1)
        read_syscalls = tables.Int64Col(dflt=-1)
        write_syscalls = tables.Int64Col(dflt=-1)
        read_bytes = tables.Int64Col(dflt=-1)
        write_bytes = tables.Int64Col(dflt=-1)
        cancelled_write_bytes = tables.Int64Col(dflt=-1)
        nvcsw = tables.Int64Col(dflt=-1)
        nivcsw = tables.Int64Col(dflt=-1)
        ac_utimescaled = tables.Int64Col(dflt=-1)
        ac_stimescaled = tables.Int64Col(dflt=-1)
        cpu_scaled_run_real_total = tables.Int64Col(dflt=-1)
        freepages_count = tables.Int64Col(dflt=-1)
        freepages_delay_total = tables.Int64Col(dflt=-1)

    missing_keys = []

    @classmethod
    def get_table_description(cls):
        return cls.HDF5TableDescription

    @classmethod
    def to_counters(cls, row):
        counters = cls.Counters()
        for key in cls.Counters._header_definitions.keys():
            counters._set_header(key, row[key])

        for key in cls.Counters._counter_definitions.keys():
            counters._set_counter(key, row[key])
        return counters

    @classmethod
    def to_row(cls, row, counters):
        for key in cls.Counters._header_definitions.keys():
            try:
                row[key] = counters._get_header(key)
            except Exception as e:
                if key not in cls.missing_keys:
                    cls.missing_keys.append(key)
                    LOG.warning(e)
        for key in cls.Counters._counter_definitions.keys():
            try:
                row[key] = counters._get_counter(key)
            except Exception as e:
                if key not in cls.missing_keys:
                    cls.missing_keys.append(key)
                    LOG.warning(e)


class HDF5ProcstatsCounters(object):
    Counters = get_counters_class("procstats_default")

    class HDF5TableDescription(tables.IsDescription):

        metric_backend = tables.StringCol(255)
        timestamp = tables.Int64Col(dflt=-1)
        hostname = tables.StringCol(255)
        job_id = tables.Int64Col(dflt=-1)

        uptime_total = tables.Int64Col(dflt=-1)
        uptime_idle = tables.Int64Col(dflt=-1)

        meminfo_memtotal = tables.Int64Col(dflt=-1)
        meminfo_memfree = tables.Int64Col(dflt=-1)
        meminfo_buffers = tables.Int64Col(dflt=-1)
        meminfo_cached = tables.Int64Col(dflt=-1)
        meminfo_swapcached = tables.Int64Col(dflt=-1)
        meminfo_active = tables.Int64Col(dflt=-1)
        meminfo_inactive = tables.Int64Col(dflt=-1)
        meminfo_unevictable = tables.Int64Col(dflt=-1)
        meminfo_mlocked = tables.Int64Col(dflt=-1)
        meminfo_swaptotal = tables.Int64Col(dflt=-1)
        meminfo_swapfree = tables.Int64Col(dflt=-1)
        meminfo_dirty = tables.Int64Col(dflt=-1)
        meminfo_writeback = tables.Int64Col(dflt=-1)
        meminfo_anonpages = tables.Int64Col(dflt=-1)
        meminfo_mapped = tables.Int64Col(dflt=-1)
        meminfo_shmem = tables.Int64Col(dflt=-1)
        meminfo_slab = tables.Int64Col(dflt=-1)
        meminfo_sreclaimable = tables.Int64Col(dflt=-1)
        meminfo_sunreclaim = tables.Int64Col(dflt=-1)
        meminfo_kernelstack = tables.Int64Col(dflt=-1)
        meminfo_pagetables = tables.Int64Col(dflt=-1)
        meminfo_nfs_unstable = tables.Int64Col(dflt=-1)
        meminfo_bounce = tables.Int64Col(dflt=-1)
        meminfo_writebacktmp = tables.Int64Col(dflt=-1)
        meminfo_commitlimit = tables.Int64Col(dflt=-1)
        meminfo_committed_as = tables.Int64Col(dflt=-1)
        meminfo_vmalloctotal = tables.Int64Col(dflt=-1)
        meminfo_vmallocused = tables.Int64Col(dflt=-1)
        meminfo_vmallocchunk = tables.Int64Col(dflt=-1)
        meminfo_hardwarecorrupted = tables.Int64Col(dflt=-1)
        meminfo_anonhugepages = tables.Int64Col(dflt=-1)
        meminfo_hugepages_total = tables.Int64Col(dflt=-1)
        meminfo_hugepages_free = tables.Int64Col(dflt=-1)
        meminfo_hugepages_rsvd = tables.Int64Col(dflt=-1)
        meminfo_hugepages_surp = tables.Int64Col(dflt=-1)
        meminfo_hugepagesize = tables.Int64Col(dflt=-1)
        meminfo_directmap4k = tables.Int64Col(dflt=-1)
        meminfo_directmap2m = tables.Int64Col(dflt=-1)

        vmstat_pgpgin = tables.Int64Col(dflt=-1)
        vmstat_pgpgout = tables.Int64Col(dflt=-1)
        vmstat_pswpin = tables.Int64Col(dflt=-1)
        vmstat_pswpout = tables.Int64Col(dflt=-1)
        vmstat_pgfault = tables.Int64Col(dflt=-1)
        vmstat_pgmajfault = tables.Int64Col(dflt=-1)

        stat_cpu_user = tables.Int64Col(dflt=-1)
        stat_cpu_nice = tables.Int64Col(dflt=-1)
        stat_cpu_system = tables.Int64Col(dflt=-1)
        stat_cpu_idle = tables.Int64Col(dflt=-1)
        stat_cpu_iowait = tables.Int64Col(dflt=-1)
        stat_cpu_irq = tables.Int64Col(dflt=-1)
        stat_cpu_softirq = tables.Int64Col(dflt=-1)
        stat_cpu_guest = tables.Int64Col(dflt=-1)
        stat_cpu_guest_nice = tables.Int64Col(dflt=-1)

        stat_intr = tables.Int64Col(dflt=-1)
        stat_ctxt = tables.Int64Col(dflt=-1)
        stat_processes = tables.Int64Col(dflt=-1)
        stat_procs_blocked = tables.Int64Col(dflt=-1)

        loadavg_1min = tables.Float32Col(dflt=-1)
        loadavg_5min = tables.Float32Col(dflt=-1)
        loadavg_15min = tables.Float32Col(dflt=-1)
        loadavg_runnable = tables.Float32Col(dflt=-1)
        loadavg_total_threads = tables.Float32Col(dflt=-1)

        # sys_numa_zoneallocs = tables.Int64Col(dflt=-1)
        # sys_numa_zoneallocs = tables.Int64Col(dflt=-1)
        # sys_numa_foreign_allocs = tables.Int64Col(dflt=-1)
        # sys_numa_allocation = tables.Int64Col(dflt=-1)
        # sys_numa_allocation = tables.Int64Col(dflt=-1)
        # sys_numa_interleave = tables.Int64Col(dflt=-1)

    missing_keys = []

    @classmethod
    def get_table_description(cls):
        return cls.HDF5TableDescription

    @classmethod
    def to_counters(cls, row):
        counters = cls.Counters()
        for key in cls.Counters._header_definitions.keys():
            counters._set_header(key, row[key])

        for key in cls.Counters._counter_definitions.keys():
            counters._set_counter(key, row[key])
        return counters

    @classmethod
    def to_row(cls, row, counters):
        for key in cls.Counters._header_definitions.keys():
            try:
                row[key] = counters._get_header(key)
            except Exception as e:
                if key not in cls.missing_keys:
                    cls.missing_keys.append(key)
                    LOG.warning(e)
        for key in cls.Counters._counter_definitions.keys():
            try:
                row[key] = counters._get_counter(key)
            except Exception as e:
                if key not in cls.missing_keys:
                    cls.missing_keys.append(key)
                    LOG.warning(e)


class HDF5OutputBackend(OutputBaseBackend):
    '''
    stdout backend class
    '''
    __backend_name__ = "hdf5"

    def open(self):
        self.stat_buffer = dict()
        self.jobs = {}

    def _get_job_stat(self, job_id):
        if job_id not in self.jobs:
            self.jobs[job_id] = JobFile(self.options, job_id)
        return self.jobs[job_id]

    def close(self):
        for job_file in self.jobs.values():
            job_file.close_job_file()

    def push(self, counters_list):
        '''
        put the metrics to the output backend
        '''
        counters_dict = dict()
        for counters in counters_list:
            job_id = counters._get_header('job_id')
            if job_id not in counters_dict:
                counters_dict[job_id] = list()
            counters_dict[job_id].append(counters)

        for (job_id, c_list) in counters_dict.iteritems():
            jobstat = self._get_job_stat(job_id)
            jobstat.append_stats(c_list)


class FileAccess(object):
    '''
    Share the access to one or several between each monitored job
    '''
    def __init__(self):
        self.hdf5_files = dict()

    def _open_hdf5_file(self, path, filemode, filters):
        """
        Open the hdf5 file corresponding to the job.
        """
        LOG.debug("Opening the file %s" % path)
        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        hdf5_file = tables.openFile(path, filemode, filters=filters)

        return hdf5_file

    def open_file(self, path, filemode, filters=None):
        if path in self.hdf5_files:
            hdf5_file = self.hdf5_files[path]
            if hdf5_file.mode != filemode:
                raise FileAlreadyOpenWithDifferentModeError(path)
        else:
            hdf5_file = self._open_hdf5_file(path, filemode, filters=filters)
            self.hdf5_files[path] = hdf5_file
        return hdf5_file

    def close_file_by_path(self, path):
        if path in self.hdf5_files:
            self.hdf5_files[path].close()


class JobFile(object):
    fileaccess = FileAccess()
    hdf5_counters = {
        HDF5ProcstatsCounters.Counters: HDF5ProcstatsCounters,
        HDF5TaskstatsCounters.Counters: HDF5TaskstatsCounters
    }
    path_level = 4

    def __init__(self, options, job_id, filemode="a"):
        self.options = options
        self.job_id = job_id
        self.job_file = None
        self.job_filemode = filemode
        self.job_table = None
        self.job_metric_counters_class = None
        self.job_metric_backend = None

        if hasattr(options, 'hdf5_filepath') \
                and options.hdf5_filepath is not None:
            self.hdf5_filepath = options.hdf5_filepath
        else:
            self.hdf5_filepath = "/tmp/colmet/hdf5/counters"

        if hasattr(options, 'hdf5_complevel'):
            self.hdf5_complevel = options.hdf5_complevel
        else:
            self.hdf5_complevel = 0
        if hasattr(options, 'hdf5_complib'):
            self.hdf5_complib = options.hdf5_complib
        else:
            self.hdf5_complib = None

        LOG.debug("Writing counters in hdf5 format for job %s"
                  % self.job_id)

    def _open_job_file(self):
        """
        Open the hdf5 file corresponding to the job.
        """
        if self.hdf5_complevel == 0 or self.hdf5_complib is None:
            self.job_file = JobFile.fileaccess.open_file(self.hdf5_filepath,
                                                         self.job_filemode)
        else:
            LOG.info("HDF5 compression enabled (lib=%s, level=%s) for %s" %
                     (self.hdf5_complib, self.hdf5_complevel,
                      self.hdf5_filepath))
            filters = tables.Filters(complevel=self.hdf5_complevel,
                                     complib=self.hdf5_complib)
            self.job_file = JobFile.fileaccess.open_file(self.hdf5_filepath,
                                                         self.job_filemode,
                                                         filters=filters)

    def close_job_file(self):
        """
        Close the hdf5 file corresponding to the job.
        """
        JobFile.fileaccess.close_file_by_path(self.hdf5_filepath)

    @property
    def job_metric_hdf5_class(self):
        return self.hdf5_counters[self.job_metric_counters_class]

    def _init_job_file_if_needed(self, metric_backend=None):
        self._open_job_file()

        group_name = "job_%s" % self.job_id
        group_path = "/%s" % group_name
        if group_path not in self.job_file:
            root = self.job_file.root
            self.job_file.createGroup(root, group_name)

        table_name = "metrics"
        table_path = "%s/%s" % (group_path, table_name)

        if table_path not in self.job_file:
            if metric_backend is None:
                raise ValueError(
                    "The metric backend must be "
                    "defined to create the table"
                )

            self.job_metric_backend = metric_backend
            self.job_metric_counters_class = get_counters_class(metric_backend)
            self.job_file.createTable(
                group_path,
                table_name,
                self.job_metric_hdf5_class.get_table_description(),
                "Metrics for the Job %s" % self.job_id
            )
            self.job_table = self.job_file.getNode(table_path)
            self.job_table.setAttr('metric_backend', metric_backend)
            self.job_table.setAttr('backend_version', HDF5_BACKEND_VERSION)

        else:
            self.job_table = self.job_file.getNode(table_path)
            self.job_metric_backend = self.job_table.getAttr('metric_backend')
            self.job_backend_version = \
                self.job_table.getAttr('backend_version')
            self.job_metric_counters_class = \
                get_counters_class(self.job_metric_backend)

    def append_stats(self, stats):
        if self.job_table is None:
            self._init_job_file_if_needed(stats[0].metric_backend)

        row = self.job_table.row
        for stat in stats:
            self.job_metric_hdf5_class.to_row(row, stat)
            row.append()

        self.job_table.flush()




import struct
import logging


LOG = logging.getLogger()

from .base import UInt8, UInt16, UInt32, UInt64, String, BaseCounters


class TaskstatsCounters(BaseCounters):
    __metric_name__ = 'taskstats_default'

    counters_taskstats = {
        # 'key' : ( offset,length, type, repr, acc, init_value )
        #
        # /* The version number of this struct. This field is always set to
        #  * TAKSTATS_VERSION, which is defined in <linux/taskstats.h>.
        #  * Each time the struct is changed, the value should be incremented.
        #  */
        # __u16   version;
        'version': (0, UInt16(), '', 'none', 'Version'),
        # __u32   ac_exitcode;            /* Exit status */
        'ac_exitcode': (2, UInt32(), '', 'none', 'Exit code'),

        # /* The accounting flags of a task as defined in <linux/acct.h>
        #  * Defined values are AFORK, ASU, ACOMPAT, ACORE, and AXSIG.
        #  */
        # __u8    ac_flag;                /* Record flags */
        'ac_flag': (6, UInt8(), '', 'none', 'Flag'),
        # __u8    ac_nice;                /* task_nice */
        'ac_nice': (7, UInt8(), '', 'none', 'Nice'),

        # /* Delay accounting fields start
        #  *
        #  * All values, until comment "Delay accounting fields end" are
        #  * available only if delay accounting is enabled, even though the
        #  * last few fields are not delays
        #  *
        #  * xxx_count is the number of delay values recorded
        #  * xxx_delay_total is the corresponding cumulative delay in
        #  * nanoseconds
        #  *
        #  * xxx_delay_total wraps around to zero on overflow
        #  * xxx_count incremented regardless of overflow
        #  */

        # /* Delay waiting for cpu, while runnable
        #  * count, delay_total NOT updated atomically
        #  */
        # __u64   cpu_count __attribute__((aligned(8)));
        'cpu_count': (16, UInt64(), 'count', 'add',
                      "Nb of cpu delay values recorded"),
        # __u64   cpu_delay_total;
        'cpu_delay_total': (24, UInt64(), 'nsec', 'add',
                            "Total of cumulative cpu delay"),

        # /* Following four fields atomically updated using task->delays->lock
        #  */

        # /* Delay waiting for synchronous block I/O to complete
        #  * does not account for delays in I/O submission
        #  */
        # __u64   blkio_count;
        'blkio_count': (32, UInt64(), 'count', 'add',
                        "Nb of block I/O delay values recorded"),
        # __u64   blkio_delay_total;
        'blkio_delay_total': (40, UInt64(), 'nsec', 'add',
                              "Total of cumulative block I/O delay"),

        # /* Delay waiting for page fault I/O (swap in only) */
        # __u64   swapin_count;
        'swapin_count': (48, UInt64(), 'count', 'add',
                         "Nb of swapin delay calues recorded"),
        # __u64   swapin_delay_total;
        'swapin_delay_total': (56, UInt64(), 'nsec', 'add',
                               "Total of cumulative swapin delay"),

        # /* cpu "wall-clock" running time
        #  * On some architectures, value will adjust for cpu time stolen
        #  * from the kernel in involuntary waits due to virtualization.
        #  * Value is cumulative, in nanoseconds, without a corresponding count
        #  * and wraps around to zero silently on overflow
        #  */
        # __u64   cpu_run_real_total;
        'cpu_run_real_total': (64, UInt64(), 'nsec', 'add',
                               "Total of cumulative cpu run real"),

        # /* cpu "virtual" running time
        #  * Uses time intervals seen by the kernel i.e. no adjustment
        #  * for kernel's involuntary waits due to virtualization.
        #  * Value is cumulative, in nanoseconds, without a corresponding count
        #  * and wraps around to zero silently on overflow
        #  */
        # __u64   cpu_run_virtual_total;
        'cpu_run_virtual_total': (72, UInt64(), 'nsec', 'add',
                                  "Total of cumulative cpu run virtual"),
        # /* Delay accounting fields end */
        # /* version 1 ends here */

        # /* Basic Accounting Fields start */
        # char    ac_comm[TS_COMM_LEN];   /* Command name */
        'ac_comm': (80, String(32), 'string', 'none', ""),
        # __u8    ac_sched __attribute__(aligned(8: 'none'));
        # /* Scheduling discipline */
        'ac_shed': (112, UInt8(), '', 'none', ""),
        # __u8    ac_pad[3];
        'ac_pad': (113, String(3), '', 'none', ""),
        # __u32   ac_uid __attribute__(aligned(8: 'none'));
        'ac_uid': (120, UInt32(), '', 'none', ""),
        # /* User ID */
        # __u32   ac_gid;                 /* Group ID */
        'ac_gid': (124, UInt32(), '', 'none', ""),
        # __u32   ac_pid;                 /* Process ID */
        'ac_pid': (128, UInt32(), '', 'none', ""),
        # __u32   ac_ppid;                /* Parent process ID */
        'ac_ppid': (132, UInt32(), '', 'none', ""),
        # __u32   ac_btime;               /* Begin time [sec since 1970] */
        'ac_btime': (136, UInt32(), 'ts_date', 'min', "Begin time"),
        # __u64   ac_etime __attribute__   (aligned(8: 'add'));
        # /* Elapsed time [usec] */
        'ac_etime': (144, UInt64(), 'usec', 'max', "Elasped time"),
        # __u64   ac_utime;               /* User CPU time [usec] */
        'ac_utime': (152, UInt64(), 'usec', 'add', "User CPU time"),
        # __u64   ac_stime;               /* SYstem CPU time [usec] */
        'ac_stime': (160, UInt64(), 'usec', 'add', "System CPU time"),
        # __u64   ac_minflt;              /* Minor Page Fault Count */
        'ac_minflt': (168, UInt64(), 'count', 'add', "Minor Page Fault Count"),
        # __u64   ac_majflt;              /* Major Page Fault Count */
        'ac_majflt': (176, UInt64(), 'count', 'add', "Major Page Fault Count"),
        # /* Basic Accounting Fields end */

        # /* Extended accounting fields start */
        # /* Accumulated RSS usage in duration of a task, in MBytes-usecs.
        #  * The current rss usage is added to this counter every time
        #  * a tick is charged to a task's system time. So, at the end we
        #  * will have memory usage multiplied by system time. Thus an
        #  * average usage per system time unit can be calculated.
        #  */
        # __u64   coremem;               /* accumulated RSS usage in MB-usec */
        'coremem': (184, UInt64(), 'mbytes-usec', 'add',
                    "Accumulated RSS usage"),
        # /* Accumulated virtual memory usage in duration of a task.
        #  * Same as acct_rss_mem1 above except that we keep track of VM usage.
        #  */
        # __u64   virtmem;               /* accumulated VM  usage in MB-usec */
        'virtmem': (192, UInt64(), 'mbytes-usec', 'add',
                    "Accumulated VM usage"),

        # /* High watermark of RSS and virtual memory usage in duration of
        #  * a task, in KBytes.
        #  */
        # __u64   hiwater_rss;         /* High-watermark of RSS usage, in KB */
        'hiwater_rss': (200, UInt64(), 'kbytes', 'max',
                        "High-watermark of RSS usage"),
        # __u64   hiwater_vm;             /* High-water VM usage, in KB */
        'hiwater_vm': (208, UInt64(), 'kbytes', 'max', "High-water VM usage"),

        # /* The following four fields are I/O statistics of a task. */
        # __u64   read_char;              /* bytes read */
        'read_char': (216, UInt64(), 'bytes', 'add', "Bytes read"),
        # __u64   write_char;             /* bytes written */
        'write_char': (224, UInt64(), 'bytes', 'add', "Bytes written"),
        # __u64   read_syscalls;          /* read syscalls */
        'read_syscalls': (232, UInt64(), 'count', 'add', "Read syscalls"),
        # __u64   write_syscalls;         /* write syscalls */
        'write_syscalls': (240, UInt64(), 'count', 'add', "Write syscalls"),
        # /* Extended accounting fields end */

        # define TASKSTATS_HAS_IO_ACCOUNTING
        # /* Per-task storage I/O accounting starts */
        # __u64   read_bytes;             /* bytes of read I/O */
        'read_bytes': (248, UInt64(), 'bytes', 'add', "Read I/O"),
        # __u64   write_bytes;            /* bytes of write I/O */
        'write_bytes': (256, UInt64(), 'bytes', 'add', "Write I/O"),
        # __u64   cancelled_write_bytes;  /* bytes of cancelled write I/O */
        'cancelled_write_bytes': (264, UInt64(), 'bytes', 'add',
                                  "Cancelled write I/O"),

        # __u64  nvcsw;                   /* voluntary_ctxt_switches */
        'nvcsw': (272, UInt64(), 'count', 'add', "Voluntary context switches"),

        # __u64  nivcsw;                  /* nonvoluntary_ctxt_switches */
        'nivcsw': (280, UInt64(), 'count', 'add',
                   "Non-volontary context switches"),
        # /* time accounting for SMT machines */
        # __u64   ac_utimescaled;         /* utime scaled on frequency etc */
        'ac_utimescaled': (288, UInt64(), 'usec', 'add',
                           "User CPU time scaled on frequency"),
        # __u64   ac_stimescaled;         /* stime scaled on frequency etc */
        'ac_stimescaled': (296, UInt64(), 'usec', 'add',
                           "System CPU time scaled on frequency"),
        # __u64   cpu_scaled_run_real_total; /* scaled cpu_run_real_total */
        'cpu_scaled_run_real_total':
            (304, UInt64(), 'nsec',
             'add', "CPU run real total scaled on frequency"),

        # /* Delay waiting for memory reclaim */
        # __u64   freepages_count;
        'freepages_count': (312, UInt64(), 'count', 'add',
                            "Number of freepages delay values recorded"),
        # __u64   freepages_delay_total;
        'freepages_delay_total': (320, UInt64(), 'nsec', 'add',
                                  "Total of cumulative freepages delay"),
    }

    counters_taskstats_to_get = [
        # 'version',
        # 'ac_exitcode',
        # 'ac_flag',
        # 'ac_nice',
        'cpu_count',
        'cpu_delay_total',
        'blkio_count',
        'blkio_delay_total',
        'swapin_count',
        'swapin_delay_total',
        'cpu_run_real_total',
        'cpu_run_virtual_total',
        # 'ac_comm',
        # 'ac_shed',
        # 'ac_pad',
        # 'ac_uid',
        # 'ac_gid',
        # 'ac_pid',
        # 'ac_ppid',
        'ac_btime',
        'ac_etime',
        'ac_utime',
        'ac_stime',
        'ac_minflt',
        'ac_majflt',
        'coremem',
        'virtmem',
        # 'hiwater_rss',
        # 'hiwater_vm',
        'read_char',
        'write_char',
        'read_syscalls',
        'write_syscalls',
        'read_bytes',
        'write_bytes',
        'cancelled_write_bytes',
        'nvcsw',
        'nivcsw',
        'ac_utimescaled',
        'ac_stimescaled',
        'cpu_scaled_run_real_total',
        'freepages_count',
        'freepages_delay_total',
    ]

    _counters = []
    for c_name in counters_taskstats_to_get:
        (_, c_type, c_repr, c_acc, c_descr) = counters_taskstats[c_name]
        _counters.append((c_name, c_type, c_repr, c_acc, c_descr))

    @classmethod
    def get_zero_counters(cls):
        return cls(taskstats_buffer=None, raw=None)

    @classmethod
    def fetch(cls, taskstat_backend, request):
        return taskstat_backend.get_task_stats(request)

    @classmethod
    def build_request(cls, taskstats_backend, tid):
        return taskstats_backend.build_request(tid)

    def __init__(self, taskstats_buffer=None, raw=None):
        BaseCounters.__init__(self, raw=raw)
        if raw is not None:
            pass
        elif taskstats_buffer is None:
            self._empty_fill()
        else:
            for name in TaskstatsCounters._counter_definitions:
                packed_values = TaskstatsCounters.counters_taskstats[name]
                (c_offset, c_type, _, _, _) = packed_values
                data = taskstats_buffer[c_offset:c_offset + c_type.length]
                unpacked_struct = struct.unpack(c_type.struct_code, data)
                self._counter_values[name] = unpacked_struct[0]



import pkgutil

__path__ = pkgutil.extend_path(__path__, __name__)  # noqa
for importer, modname, ispkg in pkgutil.walk_packages(path=__path__,
                                                      prefix=__name__ + '.'):
        __import__(modname)

from colmet.common.exceptions import UnableToFindCounterClassError
from .base import BaseCounters


def find_inheritors_counters(klass):
    subclasses = {}
    work = [klass]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses[child.__metric_name__] = child
                work.append(child)
    return subclasses

counters_registry = find_inheritors_counters(BaseCounters)


def get_counters_class(metric):
    if metric in counters_registry:
        return counters_registry[metric]
    raise UnableToFindCounterClassError(metric)



# -*- coding: utf-8 -*-
'''
    Colmet
    ------
    Collecting metrics about jobs (task/process/cgroup/etc.) running in a
    distributed environnement

'''
__version__ = '0.5.2'
VERSION = __version__
